Class: Ractor::Wrapper::StoppedError
- Defined in:
- lib/ractor/wrapper.rb
Overview
Raised when calling a method on a Ractor::Wrapper whose server has stopped and is no longer accepting calls.
Raised when calling a method on a Ractor::Wrapper whose server has stopped and is no longer accepting calls.